/*
  This is VESA video driver, same as the internal video driver in Display.exe.
  I provide the source as an example to show you how to write external video
  driver for Display.
  
  Note : You must have installed DJGPP 2.01. Please don't ask me how to
         modify this file. Study it yourself.
         Don't do any I/O in this file

  After building video.drv, copy it to the Display directory.         

  Written by Jih-Shin Ho,   1-14-1997         
*/

#include <dpmi.h>
#include <sys/farptr.h>
#include <stdlib.h>
#ifndef IS_DXE
#include <stdio.h>
#endif
#include <sys/movedata.h>
#include "videodrv.h"

#define F_COLOR_RGB 0
#define F_COLOR_BGR 1

/*
  Portions of code from VESA driver in DJGPP 1.x
*/

#define VESA_OK 0
#define VESA_ERROR 1
  
/*
 * VESA BIOS extended function number and return code
 */
#define VESA_FUNC	0x4f00
#define VESA_SUCCESS	0x004f

#define VESA_VERSION(major,minor)	((((major) & 0xff) << 8) | ((minor) & 0xff))
#define VESA_VERSION_MAJOR(vers)	(((vers) >> 8) & 0xff)
#define VESA_VERSION_MINOR(vers)	((vers) & 0xff)

/*
 * VESA BIOS sub-function numbers
 */
#define VESA_VBE_INFO	0		/* get VGA adapter info */
#define VESA_MODE_INFO  1		/* get SVGA mode info */
#define VESA_SET_MODE	2		/* set video mode */
#define VESA_GET_MODE	3		/* get current video mode */
#define VESA_VGA_STATE  4		/* save/restore VGA state */
#define VESA_PAGE_CTL	5		/* memory control */
/* VESA 1.1 +++ */
#define VESA_SCAN_LNLEN 6		/* set/get scan line length */
#define VESA_DISP_START 7		/* set/get display start address */
/* VESA 1.2 +++ */
#define VESA_PAL_CNTRL  8		/* DAC palette control */
#define VESA_PMODE      10              /* protected mode interface */


/*
 * The Vesa info structure
 */
typedef struct {
    char	    VbeSignature[4];	/* should be "VESA" */
    unsigned short  VbeVersion;		/* VESA version number */
    unsigned long   OemStringPtr;	/* Pointer to OEM string */
    unsigned long   Capabilities;	/* capabilities of the video env */
    unsigned short  VideoModeOff;	/* ptr to supported Super VGA modes */
    unsigned short  VideoModeSeg;
    /*
     * memory size in 64K blocks in VESA 1.2 and later
     */
    unsigned short    	    TotalMemory;	/* # of 64K pages */
    /*
     * Added for VBE 2.0
     */
     unsigned short OemSoftwareRev;	/* VBE implementation Software revision */
     unsigned long  OemVendorNamePtr;	/* Pointer to Vendor Name String */
     unsigned long  OemProductNamePtr;	/* Pointer to Product Name String */
     unsigned long  OemProductRevPtr;	/* Pointer to Product Revision String */
     char	    Reserved[222];	/* Reversed for VBE */
     unsigned char  OemData[256];	/* Data Area for OEM Strings */

} __attribute__ ((__packed__)) VbeInfoBlock;

#define VBE_DAC_SWITCHABLE 1
#define VBE_VGA_COMP       2


/*
 * The mode information structure
 */
typedef struct {
    unsigned short  ModeAttributes;	/* mode attributes */
    unsigned char   WinAAttributes;	/* Window A attributes */
    unsigned char   WinBAttributes;	/* Window B attributes */
    unsigned short  WinGranularity;	/* window granularity */
    unsigned short  WinSize;		/* window size */
    unsigned short  WinASegment;	/* Window A start segment */
    unsigned short  WinBSegment;	/* Window B start segment */
    unsigned long   WinFuncPtr;		/* pointer to window function */
    unsigned short  BytesPerScanLine;	/* bytes per scan line */
    /*
     * extended information for VBE 1.2 and above
     */
    short	    XResolution;	/* horizontal resolution */
    short	    YResolution;	/* vertical resolution */
    char	    XCharSize;		/* character cell width */
    char	    YCharSize;		/* character cell height */
    char	    NumberOfPlanes;	/* number of memory planes */
    char	    BitsPerPixel;	/* bits per pixel */
    char	    NumberOfBanks;	/* number of banks */
    char	    MemoryModel;	/* memory model type */
    char	    BankSize;		/* bank size in K */
    char	    NumImagePages;	/* number of image pages */
    char	    reserved[1];
    /*
     * VESA 1.2 and later
     */
    char	    RedMaskSize;	/* number of bits in red mask */
    char	    RedFieldPosition;   /* starting bit for red mask */
    char	    GreenMaskSize;
    char	    GreenFieldPosition;
    char	    BlueMaskSize;
    char	    BlueFieldPosition;
    char	    RsvdMaskSize;    	/* reserved bits in pixel */
    char	    RsvdFieldPosition;
    char	    DirectColorModeInfo; /* direct color mode attributes */
    /*
     * Mandatory information for VBE 2.0 and above
     */
    unsigned int    PhysBasePtr;	/* physical address for flat memory frame buffer */
    unsigned int    OffScreenMemOffset; /* pointer to start of off screen memory */
    unsigned short  OffScreenMemSize;	/* amount of off screen memory in 1k units */
    char 	    Reserved[206];	/* remainder of ModeInfoBlock */
    
} __attribute__ ((__packed__)) ModeInfoBlock;


/*
 * MODE attribute bits
 */
#define MODE_SUPPORTED  1		/* Mode supported in hardware */
#define MODE_EXTINFO	2		/* Extended information available */
#define MODE_SUPBIOS	4		/* Text output supported by BIOS */
#define MODE_ISCOLOR	8		/* Monochrome/color mode */
#define MODE_ISGRAPHICS 16		/* Mode type (0: text, 1:graphics) */
#define MODE_NOTVGA	32		/* 1 : Not VGA compatable */
#define MODE_NOVGAWIN	64		/* 1 : No VGA windowsing */
#define MODE_FLAT	128		/* 1 : Flat frame buffer */

/*
 * Window attribute bits
 */
#define WIN_SUPPORTED	1		/* Window supported */
#define WIN_READABLE 	2		/* Window readable */
#define WIN_WRITABLE 	4		/* Window writable */

/*
 * MemoryModel values
 */
#define MODEL_TEXT	0		/* 00h = Text mode */
#define MODEL_CGA	1		/* 01h = CGA graphics */
#define MODEL_HERC	2		/* 02h = Hercules graphics */
#define MODEL_4PLANE	3		/* 03h = 4-plane planar */
#define MODEL_PACKED	4		/* 04h = Packed pixel */
#define MODEL_256_NC	5		/* 05h = Non-chain 4, 256 color */
#define MODEL_DIRECT	6		/* 06h = direct color mode */
/* 07h-0Fh = Reserved, to be defined by VESA */
/* 10h-FFh = To be defined by OEM	     */


/*
  System variable. Don't change
*/  
static unsigned int vesa_dos_tb,vesa_dos_ds;
static int (*vesa_dpmi_int)(int, ...);
static void *(*vesa_malloc)(size_t);
static void (*vesa_free)(void *);
static void (*vesa_qsort)(void *, ...);
static void *(*vesa_bsearch)(void *, ...);


static int VesaVersion;
static int VesaTotalMemory;
static unsigned int VesaPhysBasePtr;


/* ==================================== */
static __inline__ int vesa_my_ds(void)
{
  short result;
  __asm__("movw %%ds,%0" : "=r" (result));
  return result;
}


/* ==================================== */
static int vesa_get_vbeinfoblock(VbeInfoBlock *block)
{                               
  __dpmi_regs regs;
  int vesa_got;

  vesa_got = 0;
  regs.x.es = vesa_dos_tb >> 4;
  regs.x.di = vesa_dos_tb & 0xf;
  regs.x.ax = VESA_FUNC + VESA_VBE_INFO;
  vesa_dpmi_int(0x10,&regs);
  if (regs.x.ax == VESA_SUCCESS) {
#if 1  
    movedata(vesa_dos_ds,vesa_dos_tb,vesa_my_ds(),(unsigned int)block,
             sizeof(VbeInfoBlock));
#else      
    dosmemget(vesa_dos_tb,sizeof(VbeInfoBlock),block);
#endif    
    if (block->VbeSignature[0] == 'V' && block->VbeSignature[1] == 'E' &&
        block->VbeSignature[2] == 'S' && block->VbeSignature[3] == 'A') {
      VesaVersion = block->VbeVersion;
      VesaTotalMemory = block->TotalMemory * 64;
      vesa_got = 1;
    }
  }
  return(vesa_got ? VESA_OK : VESA_ERROR);
}


/* ==================================== */
static int vesa_get_modeinfoblock(int mode,ModeInfoBlock *block)
{
  __dpmi_regs regs;
  int status;
  unsigned int tb;

  tb = vesa_dos_tb + 2048;
  regs.x.es = tb >> 4;
  regs.x.di = tb & 0xf;
  regs.x.ax = VESA_FUNC + VESA_MODE_INFO;
  regs.x.cx = mode;
  vesa_dpmi_int(0x10,&regs);
  if (regs.x.ax == VESA_SUCCESS) {
#if 1  
    movedata(vesa_dos_ds,tb,vesa_my_ds(),(unsigned int)block,
             sizeof(ModeInfoBlock));
#else
    dosmemget(tb,sizeof(ModeInfoBlock),block);
#endif    
    status = VESA_OK;
  }
  else status = VESA_ERROR;
  return(status);
}


/* ==================================== */
static int mode_is_usable(int attrib)
{
  if ((attrib & (MODE_SUPPORTED | MODE_EXTINFO | MODE_ISCOLOR)) !=
      (MODE_SUPPORTED | MODE_EXTINFO | MODE_ISCOLOR))
    return(0);
  /*
    We don't use VGA registers, so VGA compatability is not important
  */    
  if ((attrib & (/* MODE_NOTVGA | */ MODE_NOVGAWIN)) != 0)
    return(0);
  return(1);
}


#define num_def_res (sizeof(def_width) / sizeof(short))
static short def_width[] =  { 320, 640, 640, 640, 800,1024,1280};
static short def_height[] = { 200, 350, 400, 480, 600, 768,1024};

#define MAX_G_TABLE_SIZE 60
#define MAX_T_TABLE_SIZE 20

static int g_table_size = 1;
static video_mode_type g_table[MAX_G_TABLE_SIZE+1] = {
  {  320,  200,  8,  0x013,  0 }
};

static int t_table_size = 2;
static video_mode_type t_table[MAX_T_TABLE_SIZE+1] = {
  {   80,   25,  4,  0x003,  0 },
  {   80,   50,  4,  0x003,  1 }  
};

/* ==================================== */
static void add_video_mode(video_mode_type *table,int width,int height,
                           int bit,int mode)
{
  int table_size,max_table_size,i;

  table_size = (table == g_table) ? g_table_size : t_table_size;
  max_table_size = (table == g_table) ? MAX_G_TABLE_SIZE : MAX_T_TABLE_SIZE;
  if (table_size >= max_table_size) return;

  for (i = 0; i < table_size; ++i) {
    if (table[i].width == width && table[i].height == height &&
        table[i].bit == bit) break;
  }
  if (i >= table_size) {
    table[i].width = width; table[i].height = height; table[i].bit = bit;
    table[i].mode = mode; table[i].special_func = 0;
    if (table == g_table) g_table_size++;
    else t_table_size++;
  }
}


/* ==================================== */
static int mode_cmp(const video_mode_type *t1,const video_mode_type *t2)
{
  if (t1->bit < t2->bit) return(-1);
  else if (t1->bit > t2->bit) return(1);
  if (t1->width < t2->width) return(-1);
  else if (t1->width > t2->width) return(1);  
  if (t1->height < t2->height) return(-1);
  else if (t1->height > t2->height) return(1);
  return(0);
}


static void check_vesa_flat_model(void);
static void check_vesa_prot_interface(void);

/* ==================================== */
static int vesa_video_init(unsigned int dos_tb,unsigned int dos_ds,
                           int (*dpmi_int)(int, ...),
                           void *(*_malloc)(size_t), void (*_free)(void *),
                           void (*_qsort)(void *, ...),
                           void *(*_bsearch)(void *, ...))
{
  VbeInfoBlock vbe_block;
  ModeInfoBlock mode_block;
  unsigned long VideoModePtr;
  int mode,flat;

  vesa_dos_tb = dos_tb;
  vesa_dos_ds = dos_ds;
  vesa_dpmi_int = dpmi_int;
  vesa_malloc = _malloc;
  vesa_free = _free;
  vesa_qsort = _qsort;
  vesa_bsearch = _bsearch;
  if (vesa_get_vbeinfoblock(&vbe_block) != VESA_OK) return(-1);
  VideoModePtr = vbe_block.VideoModeOff + vbe_block.VideoModeSeg * 16;
  _farsetsel(vesa_dos_ds);
  flat = 0;

  while ((mode = _farnspeekw(VideoModePtr)) != 0xffff) {
    VideoModePtr += 2;
    if (vesa_get_modeinfoblock(mode,&mode_block) != VESA_OK ||
        !mode_is_usable(mode_block.ModeAttributes))
      continue;
      
    {
      int Aattr,Battr,i;
      
      for (i = 0; (1 << i) < mode_block.WinGranularity; i++);
      if ((1 << i) != mode_block.WinGranularity) continue;
      for (i = 0; (1 << i) < mode_block.WinSize; i++);
      if ((1 << i) != mode_block.WinSize) continue;

      Aattr = mode_block.WinAAttributes;
      Battr = mode_block.WinBAttributes;
      Aattr &= (Aattr & WIN_SUPPORTED) ? (WIN_READABLE | WIN_WRITABLE) : 0;
      Battr &= (Battr & WIN_SUPPORTED) ? (WIN_READABLE | WIN_WRITABLE) : 0;

#if 0
      if (Aattr != (WIN_READABLE | WIN_WRITABLE) &&
          Battr != (WIN_READABLE | WIN_WRITABLE))
        continue;
#else
      /*
        Window needs not to be both writable and readable, but the
        combination should be.
      */   
      if (((Aattr | Battr) & (WIN_READABLE | WIN_WRITABLE)) !=
          (WIN_READABLE | WIN_WRITABLE))
        continue;  
#endif
    }
      
    if (mode_block.ModeAttributes & MODE_ISGRAPHICS) {
      int pixel_bits,good_res;

      if (mode_block.NumberOfPlanes != 1) continue;
      pixel_bits = mode_block.BitsPerPixel;
      if (mode_block.BitsPerPixel == 16 && VesaVersion >= VESA_VERSION(1,2) &&
          mode_block.RsvdMaskSize)
        pixel_bits = 15;
      if (pixel_bits != 8 && pixel_bits != 15 && pixel_bits != 16 &&
          pixel_bits != 24 && pixel_bits != 32)
        continue;
      if (mode_block.XResolution >= def_width[num_def_res - 1] &&
          mode_block.YResolution >= def_height[num_def_res - 1])
        good_res = 1;
      else {
        int i;

        for (good_res = i = 0; i < num_def_res && !good_res; ++i) {
          if (mode_block.XResolution == def_width[i] &&
              mode_block.YResolution == def_height[i])
            good_res = 1;
        }
      }
      if (!good_res) continue;
      
      add_video_mode(g_table,mode_block.XResolution,mode_block.YResolution,
                     pixel_bits,mode);
      if ((mode_block.ModeAttributes & MODE_FLAT) && flat != -1) {
        if (!VesaPhysBasePtr) {
          VesaPhysBasePtr = mode_block.PhysBasePtr; flat = 1;
        }  
        else if (VesaPhysBasePtr != mode_block.PhysBasePtr) {
          VesaPhysBasePtr = 0; flat = -1;
        }  
      }
    }
    else
      if (mode_block.XResolution >= 80 && mode_block.YResolution >= 25 &&
          mode_block.BitsPerPixel == 4)
        add_video_mode(t_table,mode_block.XResolution,mode_block.YResolution,
                       4,mode);
  }
  vesa_qsort(g_table,g_table_size,sizeof(video_mode_type),mode_cmp);
  vesa_qsort(t_table,t_table_size,sizeof(video_mode_type),mode_cmp);
       
  check_vesa_prot_interface();
  check_vesa_flat_model();

  return(0);
}


/* ==================================== */
static void vesa_video_get_table(video_mode_type **graphic_mode_table,
                                 video_mode_type **text_mode_table)
{
  *graphic_mode_table = g_table;
  *text_mode_table = t_table;
}
                    

static int vesa_win_shift,vesa_gran_shift;
static int vesa_current_bank;
static int vesa_need_flat,vesa_protected_mode_switch;

static unsigned char *vesa_prot_buffer;
static unsigned char *vesa_prot_bank_func;
static unsigned short vesa_flat_selector;


/* ==================================== */
static void vesa_video_get_info(video_info_type *info)
{
  int i;

  info->major_version = VESA_VERSION_MAJOR(VesaVersion);
  i = VESA_VERSION_MINOR(VesaVersion);
  info->minor_version = (i >> 4) * 10 + (i & 0xf);
  info->memory_size = VesaTotalMemory;
  info->have_flat_model = vesa_need_flat;
  info->protected_mode_switch = vesa_protected_mode_switch;  
}


/* ==================================== */
static int vesa_get_2byte(unsigned char *addr)
{
  return((addr[1] << 8) | addr[0]);
}


/* ==================================== */
static void check_vesa_prot_interface(void)
{
  int offset;
  __dpmi_regs regs;

  regs.x.ax = VESA_FUNC | VESA_PMODE;  /* Get vesa protected mode interface */
  regs.h.bl = 0;
  vesa_dpmi_int(0x10, &regs);
  if (regs.x.ax != VESA_SUCCESS) return;

  vesa_prot_buffer = (char *)vesa_malloc(regs.x.cx);
  if (!vesa_prot_buffer) return;
  offset = (regs.x.es << 4) + regs.x.di;
#if 1  
    movedata(vesa_dos_ds,offset,vesa_my_ds(),(unsigned int)vesa_prot_buffer,
             regs.x.cx);
#else        
  dosmemget(offset,regs.x.cx,vesa_prot_buffer);
#endif  
  offset = vesa_get_2byte(vesa_prot_buffer + 6);
  if (offset) {
    while (vesa_get_2byte(vesa_prot_buffer + offset) != 0xffff) offset += 2;
    if (vesa_get_2byte(vesa_prot_buffer + offset + 2) != 0xffff) return;
    /* Need extra memory */
  }
  offset = vesa_get_2byte(vesa_prot_buffer);
  vesa_prot_bank_func = vesa_prot_buffer + offset;
}


#define VIDEO_RAM_SIZE (4 * 1024 * 1024)
static unsigned int vesa_flat_linear;

/* ==================================== */
static void check_vesa_flat_model(void)
{
  int status;
  unsigned int linear;
  __dpmi_meminfo meminfo;

  if (!VesaPhysBasePtr) return;
  meminfo.address = VesaPhysBasePtr;
  meminfo.size = VIDEO_RAM_SIZE;
  status = __dpmi_physical_address_mapping(&meminfo);
  if (status == -1) return;
  status = __dpmi_allocate_ldt_descriptors(1);
  if (status == -1) {
    __dpmi_free_physical_address_mapping(&meminfo);
    return;
  }
  __dpmi_lock_linear_region(&meminfo);
  vesa_flat_selector = status;
  vesa_flat_linear = linear = meminfo.address;
  if (__dpmi_set_segment_base_address(vesa_flat_selector,linear) == -1 ||
      __dpmi_set_segment_limit(vesa_flat_selector, linear + VIDEO_RAM_SIZE - 1) == -1) {
    __dpmi_free_physical_address_mapping(&meminfo);
    __dpmi_free_ldt_descriptor(vesa_flat_selector);
    vesa_flat_selector = 0;
  }
}


/* ==================================== */
static void vesa_video_deinit(void)
{
  if (vesa_prot_buffer) vesa_free(vesa_prot_buffer);
  vesa_prot_buffer = vesa_prot_bank_func = 0;

  if (vesa_flat_selector) {
    __dpmi_meminfo meminfo;

    if (__dpmi_get_segment_base_address(vesa_flat_selector,&meminfo.address) != -1)
      __dpmi_free_physical_address_mapping(&meminfo);
#ifndef IS_DXE      
    else printf("get vesa_flat_selector base address error!\n");
#endif    
    if (__dpmi_free_ldt_descriptor(vesa_flat_selector) == -1)
#ifndef IS_DXE    
      printf("Free vesa_flat_selector error!\n");
#else
      {;}
#endif      
    vesa_flat_selector = 0;
  }
}


/* ==================================== */
static void vesa_video_set_option(int need_flat,int protected_mode_switch)
{
  vesa_need_flat = need_flat && vesa_flat_selector;
  vesa_protected_mode_switch = protected_mode_switch && vesa_prot_bank_func;
}


/* ==================================== */
static int vesa_video_change_bank(int bank)
{
  unsigned short save_sel;
  __dpmi_regs regs;

  if (bank == vesa_current_bank) return(bank);
  vesa_current_bank = bank;
  save_sel = _fargetsel();
  /*
    In case where window is only writable, we should set another (readable)
    window to the correct bank, too.
  */
  if (vesa_protected_mode_switch) {
    __asm__ __volatile__ (
      /* Change bank */
      /* BH : 0 = set, 1 = get. BL : 0 = A, 1 = B */
      " call *%%ecx\n"
      :
      : "a" (VESA_FUNC | VESA_PAGE_CTL), "b" (0), "c" (vesa_prot_bank_func),
        "d" (bank << (vesa_win_shift - vesa_gran_shift))
      : "ax", "bx", "cx", "dx", "si", "di"
    );
    __asm__ __volatile__ (
      /* Change bank */
      /* BH : 0 = set, 1 = get. BL : 0 = A, 1 = B */
      " call *%%ecx\n"
      :
      : "a" (VESA_FUNC | VESA_PAGE_CTL), "b" (1), "c" (vesa_prot_bank_func),
        "d" (bank << (vesa_win_shift - vesa_gran_shift))
      : "ax", "bx", "cx", "dx", "si", "di"
    );
  }
  else {
    regs.x.ax = VESA_FUNC | VESA_PAGE_CTL;
    regs.x.bx = 0;
    regs.x.dx = bank << (vesa_win_shift - vesa_gran_shift);
    vesa_dpmi_int(0x10,&regs);
    regs.x.ax = VESA_FUNC | VESA_PAGE_CTL;
    regs.x.bx = 1;
    regs.x.dx = bank << (vesa_win_shift - vesa_gran_shift);
    vesa_dpmi_int(0x10,&regs);
  }
  _farsetsel(save_sel);
  return(bank);
}


/* ==================================== */
static int vesa_video_next_bank(void)
{
  return(vesa_video_change_bank(vesa_current_bank + 1));
}


/* ==================================== */
static int vesa_video_current_bank(void)
{
  return(vesa_current_bank);
}
  

/* ==================================== */
static void special_set_mode(video_mode_type *ptr)
{
  __dpmi_regs regs;
  
  regs.x.ax = 0x1112;
  regs.x.bx = 0;
  vesa_dpmi_int(0x10,&regs);
}


/* ==================================== */
static video_mode_info_type *vesa_video_set_mode(int is_graphic,int width,
                                            int height,int bit)
{
  video_mode_type key,*ptr;
  ModeInfoBlock mode_block;
  static video_mode_info_type mode_info;
  __dpmi_regs regs;
  int table_size,must_flat,i;

  key.width = width; key.height = height; key.bit = bit;
  if (is_graphic) { ptr = g_table; table_size = g_table_size; }
  else { ptr = t_table; table_size = t_table_size; key.bit = 4; }

  ptr = vesa_bsearch(&key,ptr,table_size,sizeof(video_mode_type),mode_cmp);
  if (!ptr) return(0);
  if (ptr->mode <= 0x13) {  /* Standard VGA modes */
    regs.x.ax = ptr->mode;
    vesa_dpmi_int(0x10,&regs);
    /* vesa_win = 0; */
    if (is_graphic) {
      mode_info.win_segment = 0xA000;
      mode_info.win_shift = vesa_win_shift = vesa_gran_shift = 16;
      mode_info.bytes_per_line = ptr->width;
    }
    else {
      mode_info.win_segment = 0xB800;
      mode_info.win_shift = vesa_win_shift = vesa_gran_shift = 15;
      mode_info.bytes_per_line = ptr->width * 2;
    }
    mode_info.flat_selector = 0;
    mode_info.pixel_order = F_COLOR_RGB;
    if (ptr->special_func) special_set_mode(ptr);
  }
  else {
    if (vesa_get_modeinfoblock(ptr->mode,&mode_block) != VESA_OK) return(0);
    must_flat = (mode_block.ModeAttributes & MODE_FLAT) && vesa_need_flat &&
                is_graphic;

    regs.x.ax = VESA_FUNC | VESA_SET_MODE;
    regs.x.bx = ptr->mode | (must_flat ? 0x4000 : 0);
    vesa_dpmi_int(0x10,&regs);
    if (regs.x.ax != VESA_SUCCESS) return(0);

    for (i = 0; (1 << i) < mode_block.WinGranularity; i++);
    vesa_gran_shift = i + 10;
    i = mode_block.WinAAttributes & (WIN_SUPPORTED | WIN_WRITABLE);
    if (i == (WIN_SUPPORTED | WIN_WRITABLE)) {
      mode_info.win_segment = mode_block.WinASegment; /* vesa_win = 0; */
    }
    else { mode_info.win_segment = mode_block.WinBSegment; /* vesa_win = 1; */}
    for (i = 0; (1 << i) < mode_block.WinSize; i++);
    mode_info.win_shift = vesa_win_shift = i + 10;
    mode_info.flat_selector = must_flat ? vesa_flat_selector : 0;
    mode_info.bytes_per_line = mode_block.BytesPerScanLine;
    if (mode_block.BitsPerPixel > 8)
      mode_info.pixel_order = (mode_block.BlueFieldPosition == 0) ?
                              F_COLOR_RGB : F_COLOR_BGR;
    else mode_info.pixel_order = F_COLOR_RGB;
  }  
  vesa_current_bank = -1;
  if (is_graphic) {
    if (width == 320 && height == 200 && bit == 8) vesa_current_bank = 0;
    else vesa_video_change_bank(0);
  }
  return(&mode_info);
}



/*
  Functions exported to outter worlds. For used in DXE.
  You can't change the name of 'export_video_func'
*/                
                  
video_func_type export_video_func = {
      vesa_video_init,
      vesa_video_get_table,
      vesa_video_get_info,     
      vesa_video_deinit,      
      vesa_video_set_option,
      vesa_video_change_bank,
      vesa_video_next_bank,
      vesa_video_current_bank,
      vesa_video_set_mode
};


#ifdef VESA_TEST

#include <stdio.h>
#include <go32.h>

/* ==================================== */
static void video_get_version(int *major_version,int *minor_version)
{
  int i;

  *major_version = VESA_VERSION_MAJOR(VesaVersion);
  i = VESA_VERSION_MINOR(VesaVersion);
  *minor_version = (i >> 4) * 10 + i & 0xf;
}


/* ==================================== */
static void print_mode_table(const video_mode_type *table)
{
  int i;

  for (i = 0; table[i].bit; ++i)
    printf("%6d %6d %4d : %X\n",table[i].width,table[i].height,table[i].bit,
           table[i].mode);
  printf("\n");
}


/* ==================================== */
static void test_set_mode(int is_graphic,const video_mode_type *table)
{
  video_mode_info_type *mode_info;
  int i;

  for (i = 0; table[i].bit; ++i) {
    mode_info = vesa_video_set_mode(is_graphic,table[i].width,table[i].height,
                                    table[i].bit);
    if (!mode_info) printf("Mode %X setup error\n",table[i].mode);
    else {
      printf("Mode %X setup success\n",table[i].mode);
      printf("  win_segment    = %X\n",mode_info->win_segment);
      printf("  win_shift     = %d\n",mode_info->win_shift);
      printf("  flat_selector   = %X\n",mode_info->flat_selector);
      printf("  bytes_per_line = %d\n",mode_info->bytes_per_line);
      printf("  pixel_order    = %d\n",mode_info->pixel_order);
    }
    printf("\n");
  }
}


int main()
{
  VbeInfoBlock vbe_block;
  ModeInfoBlock mode_block;
  unsigned long VideoModePtr;
  int mode,major_version,minor_version;

  if (vesa_get_vbeinfoblock(&vbe_block) != VESA_OK) {
    printf("No VESA BIOS\n");
    return(1);
  }
  video_get_version(&major_version,&minor_version);
  printf("VESA version : %d.%d\n",major_version,minor_version);
  printf("Total memory = %dK\n",vbe_block.TotalMemory * 64);
   
  /*
   * Should we check VGA compatible ?
   */
  VideoModePtr = vbe_block.VideoModeOff + vbe_block.VideoModeSeg * 16;
  _farsetsel(_dos_ds);
  while ((mode = _farnspeekw(VideoModePtr)) != 0xffff) {
    VideoModePtr += 2;
    printf("Mode %4X:\n",mode);
    if (vesa_get_modeinfoblock(mode,&mode_block) != VESA_OK)
      printf("  Get Mode Info Error\n");
    else if (!(mode_block.ModeAttributes & MODE_SUPPORTED))
      printf("  Mode not supported\n");
    else {
      printf("  Mode is %s, %s, %s, %s\n",
             (mode_block.ModeAttributes & MODE_ISCOLOR) ? "Color" : "Mono",
             (mode_block.ModeAttributes & MODE_NOTVGA) ? "Not-VGA" : "VGA",
             (mode_block.ModeAttributes & MODE_NOVGAWIN) ? "No-VGA-Win" : "VGA-Win",
             (mode_block.ModeAttributes & MODE_FLAT) ? "Flat" : "Not-Flat");
      if (!(mode_block.ModeAttributes & MODE_EXTINFO)) {
        /* I am really delighted to have this mode, but.. */
        /* what can I do with it without resolution info? */
        printf("  No Extented Info\n");
      }
      else if (mode_block.ModeAttributes & MODE_ISGRAPHICS) {
        int actual_pixel_bits;

        actual_pixel_bits = mode_block.BitsPerPixel;
        if (mode_block.BitsPerPixel == 16 && VesaVersion >= VESA_VERSION(1,2) &&
            mode_block.RsvdMaskSize)
          actual_pixel_bits = 15;
        printf("  Graphics Mode = %d x %d x %d (plane = %d)\n",mode_block.XResolution,
               mode_block.YResolution,actual_pixel_bits,mode_block.NumberOfPlanes);
        if (mode_block.ModeAttributes & MODE_FLAT)
          printf("  Flat address = %x\n",mode_block.PhysBasePtr);
      }
      else
        printf("  Text Mode = %d x %d x %d\n",mode_block.XResolution,
               mode_block.YResolution,mode_block.BitsPerPixel);
      printf("  Bytes per Line = %d\n",mode_block.BytesPerScanLine);
    }
    printf("\n");
  }

  {
    video_mode_type *graphic_mode_table,*text_mode_table;

    printf("\n\n");
    if (vesa_video_init(__tb,_dos_ds,__dpmi_int,malloc,free,qsort,bsearch) != 0) {
      printf("Build mode table error\n");
      exit(1);
    }
    vesa_video_get_table(&graphic_mode_table,&text_mode_table);
    print_mode_table(graphic_mode_table);
    print_mode_table(text_mode_table);
  }

  printf("\n\n");
  vesa_video_set_option(1,1);
  test_set_mode(1,g_table);
  test_set_mode(0,t_table);
  vesa_video_set_mode(0,80,25,4);

  printf("\n\n");
  printf("VesaPhysBasePtr = %X\n",VesaPhysBasePtr);
  printf("vesa_flat_linear = %X\n",vesa_flat_linear);
  printf("vesa_flat_selector = %X\n",vesa_flat_selector);
  printf("vesa_prot_bank_func = %X\n",vesa_prot_bank_func);
  printf("\n\n");
  
  {
    int i,j,offset;
    video_mode_info_type *info;
    
    info = vesa_video_set_mode(1,640,480,15);
    if (!info) { vesa_video_set_mode(0,80,25,4); printf("set mode error\n"); }
    else {
      _farsetsel(vesa_flat_selector);

      for (offset = i = 0; i < 480; ++i) {
        for (j = 0; j < 640; ++j, offset += 2)
          _farnspokew(offset,0xffff);
      }
      getkey();
    }  
  }

  vesa_video_set_mode(0,80,25,4);    
  vesa_video_deinit();
  return(0);
}

#endif
